//  Copyright 2009 Stanislav Zidek
/*
 *  This file is part of CE-solver.
 *
 *  CE-solver is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CE-solver is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CE-solver.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

//Analyzer - class for analyzing dominance of strategies in games

#ifndef _ANALYZER_HH_
#define _ANALYZER_HH_
#include "analyzer.hh"

#include "game-adt.hh"
#include "ucache.hh"
#include "debug.hh"

#ifdef PARALLEL
#include <omp.h>
#endif

#include <sstream>
#include <string>
#include <vector>
#include <algorithm>
#include <stdexcept>

template <bool B, class T=float>
class Analyzer {
protected:
	enum DQType { WEAK, STRICT, ANALYZE };
	struct DominanceQuery {
		DQType type;
		std::vector<unsigned> players;
		std::vector<unsigned> strategies1;
		std::vector<unsigned> strategies2;
		DominanceQuery(DQType t,
			const std::vector<unsigned> p,
			const std::vector<unsigned> s1,
			const std::vector<unsigned> s2):
			type(t),
			players(p),
			strategies1(s1),
			strategies2(s2)
		{}
	};
	const Game<T>& game;
	UCache<B, T> ucache;
	const StrategyProfile strategies;
public:
	Analyzer(const Game<T>& g):
		game(g),
		ucache(g),
		strategies(game.strategies())
	{}

	std::string query(std::string input)
	{
		unsigned col_count = std::count(input.begin(), input.end(), ':');
		if (col_count != 2) throw std::runtime_error("Query must contain two colons.");
		//assert(input.size() >= 5);
		std::string s;
		unsigned i=0;
		char c = input[i++];
		DQType t;
		if (c=='s') t = STRICT;
		else if (c=='w') t = WEAK;
		else if (c=='a') t = ANALYZE;
		else {
			t = ANALYZE;
			i--;
		}

		s = input;
		s.erase(0, i);
		s.erase(s.find_first_of(':'));
		if (s.size() < 1) throw std::runtime_error("No player in query.");
		DPRINT("First s: " << s);
		std::vector<unsigned> p = srange_to_vec(s);

		s = input;
		s.erase(s.find_last_of(':'));
		s.erase(0, s.find_first_of(':'));
		if (s.size() < 2) throw std::runtime_error("No first strategy in query.");
		s.erase(0, 1);
		DPRINT("Second s: " << s);
		std::vector<unsigned> s1 = srange_to_vec(s);

		s = input;
		s.erase(0, s.find_last_of(':'));
		if (s.size() < 2) throw std::runtime_error("No second strategy in query.");
		s.erase(0, 1);
		DPRINT("Third s: " << s);
		std::vector<unsigned> s2 = srange_to_vec(s);

		std::vector<std::string> vs = dominance_analysis(DominanceQuery(t, p, s1, s2));
		std::string res;
		for (unsigned i=0; i<vs.size(); i++) {
			res += vs[i];
			res += "\n";
		}

		return res;
	}

protected:
	std::vector<unsigned> srange_to_vec(std::string s)
	{
		std::vector<unsigned> res;
		if (s == "x") return res;
		std::istringstream iss(s);
		unsigned first;
		if (!(iss >> first)) throw std::runtime_error("Bad range (first num).");
		char c = 'X';
		iss >> c;
		if (iss.eof()) {
			res.push_back(first);
		} else if (c == '-') {
			unsigned last;
			if (!(iss >> last)) throw std::runtime_error("Bad range (last num).");
			iss >> c;
			if (!iss.eof()) throw std::runtime_error("Extra characters in range.");
			if (last < first) throw std::runtime_error("Last < first.");
			for (unsigned i=first; i<=last; i++) {
				res.push_back(i);
			}
		} else if (c == ',') {
			res.push_back(first);
			do {
				bool ok = (iss >> first);
				if (ok) res.push_back(first);
				iss >> c;
				if (!iss.eof()) ok &= (c == ',');
				if (!ok) throw std::runtime_error("Error reading enumeration.");
			} while (!iss.eof());
		} else {
			throw std::runtime_error("Bad range character.");
		}

		return res;
	}

	std::vector<std::string> dominance_analysis(DominanceQuery dq)
	{
		std::vector<std::string> res;

		fill(dq);

		//std::vector<unsigned> p_v = strip(game.players, dq.players);
		int p_s = dq.players.size();
		unsigned p_m = game.players();
		for (int p_i=0; p_i<p_s; p_i++) {
			unsigned player = dq.players[p_i];
			if (player >= p_m) continue;
			int s1_s = dq.strategies1.size();
			unsigned s_m = game.strategies(player);
			for (int s1_i=0; s1_i<s1_s; s1_i++) {
				unsigned strat1 = dq.strategies1[s1_i];
				if (strat1 >= s_m) continue;
				int s2_s = dq.strategies2.size();
				for (int s2_i=0; s2_i<s2_s; s2_i++) {
					unsigned strat2 = dq.strategies2[s2_i];
					if (strat2 >= s_m) continue;
					if (strat1 == strat2) continue;
					unsigned lt=0, eq=0, gt=0;
					bool weak_possible = false;
					bool end = false;
					StrategyProfile sp(p_m);
					do {
						T diff = difference(sp, player, strat1, strat2);
						if (dq.type == STRICT) {
							if (!(diff > 0)) end = true;
						} else if (dq.type==WEAK) {
							if (diff < 0) end = true;
							weak_possible = true;
						} else if (dq.type == ANALYZE) {
							lt += diff < 0 ? 1 : 0;
							eq += diff == 0 ? 1 : 0;
							gt += diff > 0 ? 1 : 0;
						}
					} while (!end && next_profile(player, sp));
					if (!end) {
						std::ostringstream oss;
						oss << player<<"-"<<strat1<<","<<strat2<<" ";
						if (dq.type == ANALYZE) {
							oss<<"\t>"<<gt<<" ="<<eq<<" <"<<lt;
						}

						res.push_back(oss.str());
					}
				}
			}
		}

		return res;
	}

	T difference(StrategyProfile sp, unsigned player, unsigned strat1, unsigned strat2)
	{
		sp[player] = strat1;
		T u1 = ucache.get(player, sp);
		DVAR(u1);
		sp[player] = strat2;
		T u2 = ucache.get(player, sp);
		DVAR(u2);

		DVAR(u1 - u2);
		return u1 - u2;
	}

	void fill(DominanceQuery& dq)
	{
		if (dq.players.size() == 0) {
			for (unsigned i=0; i<game.players(); i++) {
				dq.players.push_back(i);
			}
		}
		StrategyProfile sp = game.strategies();
		unsigned max_strats = 0;
		for (unsigned i=0; i<sp.size(); i++) {
			if (sp[i] > max_strats) max_strats = sp[i];
		}
		if (dq.strategies1.size() == 0) {
			for (unsigned i=0; i<max_strats; i++) {
				dq.strategies1.push_back(i);
			}
		}
		if (dq.strategies2.size() == 0) {
			for (unsigned i=0; i<max_strats; i++) {
				dq.strategies2.push_back(i);
			}
		}
	}

	bool next_profile(unsigned player, StrategyProfile& prof)
	{
		DPRINT("Next1: " << prof << "/" << player);
		bool res = false;
		for (unsigned i=0; i<prof.size(); i++) {
			if (i==player) continue;
			prof[i] += 1;
			if (prof[i] >= strategies[i]) {
				prof[i] = 0;
				continue;
			}
			//normal step
			res = true;
		}
		DPRINT("Next2: " << prof << "/" << res);

		return res;
	}

	/*
	std::vector<unsigned> strip(unsigned bound, const std::vector<unsigned>& vec)
	{
		if (bound == 0) throw std::runtime_error("Zero bound.");
		std::vector<unsigned> res;
		if (vec.size() == 0) {
			res = std::vector<unsigned>(bound);
			for (unsigned i=0; i<bound; i++) {
				res[i] = i;
			}
		} else {
			for (unsigned i=0; i<vec.size(); i++) {
				if (vec[i] < bound) {
					res.push_back(vec[i]);
				}
			}
		}

		return res;
	}
	*/
};

#endif //_ANALYZER_HH_
