//  Copyright 2009 Stanislav Zidek
/*
 *  This file is part of CE-solver.
 *
 *  CE-solver is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CE-solver is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CE-solver.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

//LP problem for correlated equilibrium

#ifndef _CE_LP_HH_
#define _CE_LP_HH_
#include "ce-lp.hh"

#include "game-adt.hh"

#include <stdexcept>
#include <algorithm>
#include <vector>
#include <map>

template <class T>
class CorrelatedLP {
	//one item of constraint matrix
	struct ConstraintItem {
		unsigned row;
		unsigned col;
		T val;
		ConstraintItem(unsigned r, unsigned c, T v):row(r),col(c),val(v) {}
		inline bool operator<(const ConstraintItem& other) const
		{
			return row!=other.row ? row<other.row : col<other.col;
		}
	};
protected:
	unsigned last_index, act_row;
	std::map<StrategyProfile, unsigned> profs_vars;
	std::vector<StrategyProfile> vars_profs;
	std::vector<ConstraintItem> constraints;
	std::vector<T> obj_func;

	unsigned rows, cols, items;
public:
	CorrelatedLP ():
		last_index(0),
		act_row(0)
	{}

	//adds one constraint row (of G-matrix)
	void add_row(std::map<StrategyProfile, T> row)
	{
		DFUNC;
		typename std::map<StrategyProfile, T>::iterator iter;
		for (iter=row.begin(); iter!=row.end(); iter++) {
			T value = iter->second;
			//if (value == 0.0) continue;
			StrategyProfile prof = iter->first;
			DPRINT(prof << ": " << value);
			unsigned column = get_index(prof);
			constraints.push_back(ConstraintItem(act_row, column, value));
		}
		act_row++;

		return;
	}

	//set objective function with all coefficients =1.0
	void set_obj_func()
	{
		obj_func.clear();
		for (unsigned i=0; i<cols; i++) {
		       obj_func.push_back(1.0);
		}

		return;
	}

	//sets objective function according to "obj_map"
	void set_obj_func(std::map<StrategyProfile, T> obj_map)
	{
		assert(obj_map.size()==cols);
		obj_func.clear();
		obj_func = std::vector<T>(cols);
		typename std::map<StrategyProfile, T>::iterator it;
		for (it=obj_map.begin(); it!=obj_map.end(); it++) {
			unsigned index = get_index(it->first);
			DPRINT("SOF: "<<index<<it->first<<" -> "<<it->second);
			obj_func[index] = it->second;
		}

		return;
	}

	//constructs problem - prepares it to be solved
	void construct()
	{
		DFUNC;
		rows = act_row;
		DVAR(rows);
		cols = last_index;
		DVAR(cols);
		items = constraints.size();
		DVAR(items);
		sort();

		for (unsigned i=0; i<items; i++) {
			ConstraintItem ci = constraints[i];
			DPRINT("("<<ci.row<<", "<<ci.col<<") = "<<ci.val<<";");
		}

		return;
	}

	//gets number of rows (constraints)
	unsigned get_rows() const { return rows; }
	//gets number of columns (variables)
	unsigned get_cols() const { return cols; }
	//gets number of nonzero constraint items
	unsigned get_items() const { return items; }
	//gets objective function
	std::vector<T> get_obj_func() const { return obj_func; }

	//gets one constraint item
	void get(unsigned index, unsigned& row, unsigned& col, T& value) const
	{
		if (index >= items) {
			throw std::out_of_range("Requested constraint item out of range");
		}
		row = constraints[index].row;
		col = constraints[index].col;
		value = constraints[index].val;

		return;
	}

	//translates solution (variable indexes) to CE (strategy profiles)
	CorrelatedEquilibrium translate(const std::vector<double>& vec) const
	{
		CorrelatedEquilibrium res;

		for (unsigned i=0; i<vec.size(); i++) {
			double value = vec[i];
			if (value != 0.0) {
				res[vars_profs[i]] = value;
			}
		}

		return res;
	}

	//translates solution (variable indexes) to CE (strategy profiles)
	CorrelatedEquilibrium translate(const std::map<unsigned, double>& s) const
	{
		CorrelatedEquilibrium res;
		typename std::map<unsigned, double>::const_iterator it;

		for (it = s.begin(); it != s.end(); it++) {
			unsigned index = it->first;
			double value = it->second;
			if (value != 0.0) {
				res[vars_profs[index]] = value;
			}
		}

		return res;
	}

protected:
	//sorts constraint items
	void sort()
	{
		std::sort(constraints.begin(), constraints.end());

		return;
	}

	//gets/adds index of strategy profile
	unsigned get_index(const StrategyProfile& prof)
	{
		DFUNC;
		unsigned index;
		if (profs_vars.count(prof)) {
			DPRINT("Profile " << prof << " present.");
			index = profs_vars[prof];
		} else {
			DPRINT("Profile " << prof << " added.");
			index = last_index++;
			profs_vars[prof] = index;
			vars_profs.push_back(prof);
		}

		return index;
	}
};

#endif //_CE_LP_HH_
