//  Copyright 2009 Stanislav Zidek
/*
 *  This file is part of CE-solver.
 *
 *  CE-solver is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CE-solver is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CE-solver.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

//Miscellaneous classes for G-matrix

#ifndef _CE_MISC_HH_
#define _CE_MISC_HH_
#include "ce-misc.hh"

#include "debug.hh"
#include "game-adt.hh"

#ifdef PARALLEL
#include <omp.h>
#endif

#include <algorithm>
#include <set>
#include <vector>
#include <map>
#include <stdexcept>

//class for storing positions in G-matrix rows
class GRows {
	//structure identifying single G-matrix row
	struct RowID {
		unsigned p;
		unsigned f;
		unsigned t;
		inline bool operator<(const RowID& other) const
		{
			if (p != other.p) return (p < other.p);
			else if (f != other.f) return (f < other.f);
			else return (t < other.t);
		}
	};
protected:
	const unsigned players;
	std::map<RowID, StrategyProfile> data;
public:
	//constructor
	GRows(const unsigned n): players(n) {}

	//returns profile for given row; if no entry in database,
	//creates profile [0, ..., from, ..., 0]
	//IMPORTANT: does not check if from/to is valid strategy
	StrategyProfile get(unsigned player, unsigned from, unsigned to)
	{
		RowID id;
		id.p = player;
		id.f = from;
		id.t = to;
		StrategyProfile sp(players);
#ifdef PARALLEL
#pragma omp critical(grows)
{
#endif
		if (data.count(id) == 0) { //no such profile
			for (unsigned i=0; i<sp.size(); i++) {
				sp[i] = 0;
			}
			sp[player] = from;
		} else {
			sp = data[id];
		}
#ifdef PARALLEL
}
#endif

		return sp;
	}

	//stores profile for specified row
	void set(StrategyProfile sp, unsigned player, unsigned from, unsigned to)
	{
		RowID id;
		id.p = player;
		id.f = from;
		id.t = to;

#ifdef PARALLEL
#pragma omp critical(grows)
		data[id] = sp;
#endif

		return;
	}

	//prints all data (debugging purposes)
	void print()
	{
		std::map<RowID, StrategyProfile>::iterator it;
		for (it=data.begin(); it != data.end(); it++) {
			RowID id = it->first;
			PRINTLN(id.p << ":" << id.f << "->" << id.t << it->second);
		}

		return;
	}
};

//class storing eliminated strategies
class ForbiddenStrategies {

friend std::ostream& operator<<(std::ostream& output, const ForbiddenStrategies& fs);

protected:
	std::vector<bool> data[MAX_PLAYERS];
	const unsigned players;

public:
	ForbiddenStrategies(const unsigned p, const StrategyProfile strats):
		players(p)
	{
		for (unsigned i=0; i<players; i++) {
			data[i] = std::vector<bool>(strats[i]);
		}
	}

	//returns true if given strategy was already eliminated
	inline bool forbidden(unsigned player, unsigned index) const
	{
		assert(player<players);
		//DPRINT("Player: " << player << ", index: " << index << ":: " << data[player][index]);
		bool res;
#ifdef PARALLEL
		#pragma omp critical(forbidden_strategies)
#endif
		res = data[player][index];

		return res;
	}

	//returns true if given strategy was NOT eliminated yet
	inline bool valid(unsigned player, unsigned index) const { return !forbidden(player, index); }

	inline void add(unsigned player, unsigned index)
	{
#ifdef PARALLEL
		#pragma omp critical(forbidden_strategies)
#endif
		data[player][index] = true;
	}

	//returns number of non-eliminated strategies of given player
	//IMPORTANT: we suppose num_valid is not called in parallel section
	unsigned num_valid(unsigned player) {
		unsigned res;
		res = std::count(data[player].begin(), data[player].end(), false);

		return res;
	}

};

std::ostream& operator<<(std::ostream& output, const ForbiddenStrategies& fs)
{
	assert(fs.players>=1);
	
	output << "(";
	for (unsigned i=0; i<fs.data[0].size(); i++) {
		output << fs.data[0][i];
	}
	for (unsigned p=1; p<fs.players; p++) {
		output << ",";
		for (unsigned i=0; i<fs.data[p].size(); i++) {
			output << fs.data[p][i];
		}
	}
	output << ")";

	return output;
}

#endif //_CE_MISC_HH_
