//  Copyright 2009 Stanislav Zidek
/*
 *  This file is part of CE-solver.
 *
 *  CE-solver is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CE-solver is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CE-solver.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

//Interface for reading and solving games in files

#ifndef _FILE_IFACE_HH_
#define _FILE_IFACE_HH_
#include "file-iface.hh"

#include "debug.hh"
#include "game-adt.hh"
#include "ce-solver.hh"

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cctype>
#include <stdexcept>
#include <vector>
#include <string>

#ifndef CACHE
#define CACHE false
#endif

//for output purposes
#define THRESHOLD 1e-2 //profiles with P<THRESHOLD won't be printed
#define PRECISION 2 //how many digits to the right of the decimal point

//class inherited from Game, represents game read from file
template <class T=float>
class FileGame: public Game<T> {
private:
	unsigned num_players;
	StrategyProfile player_strats;
	std::vector<UProfile<T> > data;
	std::vector<size_t> dims;
	bool inited;

	//labelling data
	bool use_strategy_names;
	std::string game_name;
	std::vector<std::string> player_names;
	std::vector<std::vector<std::string> >strategy_names;
	std::string game_comment;
public:
	FileGame():
		inited(false),
		use_strategy_names(false),
		game_name(""),
		game_comment("")
	{}

	//reads game from input stream - recognizes input format
	//(standard, Gambit)
	void read(std::istream& is)
	{
		char c = get_char(is);
		if (c == 'N') {
			is.unget();
			read_nfg(is);
		} else if (isdigit(c)) {
			is.unget();
			read_simple(is);
		} else {
			throw std::runtime_error("Unknown file format.");
		}
		inited = true;
	}

	//returns number of players
	virtual unsigned players() const
	{
		return num_players;
	}

	//returns number of strategies of player "player"
	virtual unsigned strategies(unsigned player) const
	{
		return player_strats[player];
	}

	//returns number of strategies of all players
	virtual StrategyProfile strategies() const
	{
		return player_strats;
	}

	//returns payoff of player "p" in strategy profile "sp"
	virtual T get(const unsigned p, const StrategyProfile& sp) const
	{
		UProfile<T> up = get(sp);

		return up[p];
	}

	//returns payoff profile for strategy profile "sp"
	virtual UProfile<T> get(const StrategyProfile& sp) const
	{
		size_t index = 0;
		for (unsigned i=0; i<num_players; i++) {
			index += sp[i]*dims[i];
		}
		UProfile<T> up = data[index];
#if (DELAY>0) //computing simulation
		double iterations = DELAY * 225;
		for (double d=0; d<iterations; d += 1);
#endif
		//UProfile<T> up = (data.find(sp))->second;

		return up;
	}
protected:
	//inits data structures for storing payoffs
	void init_data()
	{
		size_t size = 1;
		for (unsigned i=0; i<num_players; i++) {
			size *= player_strats[i];
		}
		data = std::vector<UProfile<T> >(size);
	}

	//following functions parse various input files
	
	char get_char(std::istream& is)
	{
		char c;
		is >> c; //skips ws
		if (!is.good()) {
			throw std::runtime_error("Error reading input.");
		}

		return c;
	}

	T get_payoff(std::istream& is)
	{
		T tmp;
		if (!(is >> tmp)) {
			throw std::runtime_error("Cannot read payoff.");
		}
		char c;
		is >> c;
		if (is.eof()) return tmp;

		if (c == '/') {
			T denom;
			if (!(is >> denom)) {
				throw std::runtime_error("Cannot read payoff.");
			}
			tmp /= denom;
		} else {
			is.unget();
		}

		return tmp;
	}

	std::string get_string(std::istream& is)
	{
		std::string s = "";
		char c = get_char(is);
		if (c != '\"') throw std::runtime_error("Expecting double quote (string).");
		s = "";
		while ((c = get_char(is)) != '\"') {
			if (c == '\\') {
				c = is.get();
			}
			s += c;
		}

		return s;
	}

	void read_payoffs(std::istream& is)
	{
		init_data();

		UProfile<T> up(num_players);
		for (size_t index=0; index<data.size(); index++) {
			for (unsigned i=0; i<num_players; i++) {
				up[i] = get_payoff(is);
			}
			assert(index >= 0 && index < data.size());
			data[index] = up;
		}
	}

	void read_outcomes(std::istream& is)
	{
		init_data();

		std::vector<UProfile<T> > outcomes;
		char c;
		
		while (1) {
			c = get_char(is);
			if (c == '}') {
				break;
			}
			if (c != '{') {
				throw std::runtime_error("Expected set ({) - outcome.");
			}

			get_string(is); //outcome comment - ignoring
			UProfile<T> up(num_players);
			for (unsigned i=0; i<num_players; i++) {
				char c = get_char(is);
				if (c != ',') { //some files have commas
					is.unget();
				}
				up[i] = get_payoff(is);
			}
			outcomes.push_back(up);

			c = get_char(is);
			if (c != '}') {
				throw std::runtime_error("Expected end of outcome (}).");
			}
		}

		for (size_t index = 0; index < data.size(); index++) {
			size_t i;
			if (!(is >> i)) {
				throw std::runtime_error("Expected outcome index.");
			}
			if (i > outcomes.size()) {
				throw std::runtime_error("Outcome index out of range.");
			}
			data[index] = outcomes[i];
		}
	}

	//reads native file format
	void read_simple(std::istream& is)
	{
		is >> num_players;
		for (unsigned i=0; i<num_players; i++) {
			unsigned n;
			is >> n;
			player_strats[i] = n;
		}

		size_t act = 1;
		dims = std::vector<size_t>(num_players);
		for (int i=num_players; i>0; i--) {
			dims[i-1] = act;
			act *= player_strats[i-1];
		}

		read_payoffs(is);
	}

	//reads Gambit file format
	void read_nfg(std::istream& is)
	{
		read_header(is);
		read_players(is);
		read_strategies(is);
		read_comment(is);

		char c = get_char(is);
		if (c == '{') {
			read_outcomes(is);
		} else if (isdigit(c)) {
			is.unget();
			read_payoffs(is);
		} else {
			throw std::runtime_error("Expecting payoffs or outcomes");
		}

		size_t act = 1;
		dims = std::vector<size_t>(num_players);
		for (unsigned i=0; i<num_players; i++) {
			dims[i] = act;
			act *= player_strats[i];
		}
		
	}

	void read_header(std::istream& is)
	{
		std::string s;

		is >> s;
		if (s != "NFG") throw std::runtime_error("Bad file format.");
		is >> s;
		if (s != "1") throw std::runtime_error("Unknown format version (use 1).");
		is >> s;
		if (s != "R") throw std::runtime_error("Unknown numeric format (use R).");

		s = get_string(is);
		game_name = s;
	}

	void read_players(std::istream& is)
	{
		std::string s;

		is >> s;
		if (s != "{") throw std::runtime_error("Expected set of players ({).");
		bool end = false;
		while (!end) {
			char c = get_char(is);;
			if (c == '\"') {
				is.unget();
				player_names.push_back(get_string(is));
			} else if (c == '}') {
				end = true;
			} else {
				throw std::runtime_error("Unexpected character.");
			}
		}
		num_players = player_names.size();

	}

	void read_strategies(std::istream& is)
	{
		std::string s;
		std::vector<unsigned> ns;

		char c = get_char(is);
		if (c != '{') throw std::runtime_error("Expected set ({).");
		c = get_char(is);
		is.unget();
		if (isdigit(c)) { //strategy counts
			//use_strategy_names = false;
			bool end = false;
			while (!end) {
				c = get_char(is);
				if (isdigit(c)) {
					is.unget();
					unsigned num;
					is >> num;
					ns.push_back(num);
				} else if (c == '}') {
					end = true;
				} else {
					throw std::runtime_error("Unexpected character.");
				}
			}
		} else if (c == '{') { //strategy names
			use_strategy_names = true;
			while (1) {
				std::vector<std::string> names;
				c = get_char(is);
				if (c == '}') { //end
					break;
				}
				if (c != '{') {
					throw std::runtime_error("Expected set of strategy names.");
				}
				bool end = false;
				do {
					c = get_char(is);
					if (c == '}') { //end of player's strategies
						strategy_names.push_back(names);
						ns.push_back(names.size());
						names.clear(); //just for sure
						end = true;
					} else if (c == '\"') { //next strategy
						is.unget();
						s = get_string(is);
						names.push_back(s);
					} else {
						throw std::runtime_error("Expected strategy name or }.");
					}
				} while (!end);
			}
		} else {
			throw std::runtime_error("Cannot read strategies.");
		}
		if (ns.size() != num_players) {
			throw std::runtime_error("Bad strategy counts or players\' names.");
		}
		player_strats = StrategyProfile(num_players);
		for (unsigned i=0; i<num_players; i++) {
			player_strats[i] = ns[i];
		}

	}

	void read_comment(std::istream& is)
	{
		char c = get_char(is);
		is.unget();
		if (c == '\"') {
			game_comment = get_string(is);
		}
	}

	bool next(StrategyProfile& prof)
	{
		bool step=false;
		unsigned i=num_players-1;

		while (!step) {
			prof[i]++;
			if (prof[i] < player_strats[i]) {
				step = true;
				break;
			}
			prof[i] = 0;
			if (i == 0) {
				return false;
			}
			i--;
		}

		return true;
	}
};

//class solving file game - creates CESolver and solves FileGame
template <class T>
class FileSolver {
private:
	std::string filename;
	std::string reduced;
	CorrelatedEquilibrium solution;
	bool do_export;
	bool solve_lp;
public:
	//constructor - stores file name and requested actions
	FileSolver(std::string file=std::string(""), std::string red=std::string(""),
			bool de=true, bool ds=true):
		filename(file),
		do_export(de),
		solve_lp(ds)
	{}

	//solves game
	void solve()
	{
#ifdef DELAY
		PRINTLN("Delay: " << DELAY);
#endif
		FileGame<T> fg;
		if (filename != "") {
			std::ifstream ifs(filename.c_str());
			fg.read(ifs);
			ifs.close();
			//fg.read(std::ifstream(filename.c_str()));
		} else {
			fg.read(std::cin);
		}
		PRINTLN("File read.");

		CESolver<CACHE> ces(fg);
		ces.solve(do_export, solve_lp);

		if (solve_lp) {
			PRINTLN("Problem solved.");
			solution = ces.get_solution();
			threshold();
			UProfile<T> up = ces.get_payoffs();

			typename CorrelatedEquilibrium::iterator it;
			std::cout << std::fixed << std::setprecision(PRECISION);
			for (it=solution.begin(); it!=solution.end(); it++) {
				PRINTLN(it->first << " with probability " << it->second);
			}
			PRINTLN("Equilibrium payoffs: ");
			PRINTLN(up);
		}

		return;
	}

	//removes profiles with probabilities < THRESHOLD from solution
	void threshold(double thres=THRESHOLD)
	{
		CorrelatedEquilibrium sol;
		typename CorrelatedEquilibrium::iterator it;
		for (it=solution.begin(); it!=solution.end(); it++) {
			if (it->second >= thres) {
				sol[it->first] = it->second;
			}
		}
		solution = sol;
	}
};

#endif //_FILE_IFACE_HH_
