//  Copyright 2009 Stanislav Zidek
/*
 *  This file is part of CE-solver.
 *
 *  CE-solver is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CE-solver is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CE-solver.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _G_MATRIX_HH_
#define _G_MATRIX_HH_
#include "g-matrix.hh"

#include "debug.hh"
#include "print.hh"
#include "game-adt.hh"
#include "ce-lp.hh"
#include "ce-misc.hh"
#include "ucache.hh"

#ifdef PARALLEL
#include <omp.h>
#endif

#include <stdexcept>
#include <map>
#include <queue>

#include <fstream>
#include <iostream>
#include <iomanip>

template <bool B, class T>
class GMatrix {
private:
	const Game<T> &game; //reference to game model
	UCache<B, T> ucache; //payoff cache
	ForbiddenStrategies fs; //forbidden strategies
	StrategyProfile strategies; //numbers of players' strategies
	GRows rows; //positions in G-matrix

public:
	//constructs G-matrix from game model
	GMatrix(const Game<T>& g):
		game(g),
		ucache(g),
		fs(g.players(), g.strategies()),
		strategies(g.strategies()),
		rows(g.players())
	{}

	//extracts LP problem of minimized gama
	CorrelatedLP<T> get_lp_problem()
	{
		CorrelatedLP<T> lp;

		rows.print();

		bool simple = true;
		for (unsigned i=0; i<game.players(); i++) {
			if (fs.num_valid(i) != 1) {
				simple = false;
				break;
			}
		}

		//just one remaining strategy profile
		if (simple) {
			std::map<StrategyProfile, T> row;
			StrategyProfile prof(game.players());

			bool res = valid_profile(prof, game.players());
			if (!res) throw std::logic_error("No valid strategy profile.");
			row[prof] = 1;

			lp.add_row(row);
			lp.construct();
			lp.set_obj_func();

			return lp;
		}

		//more remaining strategy profiles
		std::map<StrategyProfile, T> used_profs;
		for (unsigned p=0; p<game.players(); p++) {
			unsigned pstrats = strategies[p];
			for (unsigned sfrom=0; sfrom<pstrats; sfrom++) {
				if (fs.forbidden(p, sfrom)) continue;
				for (unsigned sto=0; sto<pstrats; sto++) {
					//maybe not neccessary
					if (fs.forbidden(p, sto)) continue;
					if (sfrom != sto) {
						std::map<StrategyProfile, T> row;
						StrategyProfile prof(game.players());
						prof[p] = sfrom;

						assert(prof[p] == sfrom);
						if (!valid_profile(prof, p)) {
							throw std::logic_error("Weird row.");
						}

						do {
							T val = difference(prof, p, sfrom, sto);
							row[prof] = val;
							used_profs[prof] = 0.0;
						} while (next_profile(prof, p));

						lp.add_row(row);
					}
				}
			}
		}

		//constructs LP problem
		lp.construct();

		//sets objective function
		typename std::map<StrategyProfile, T>::iterator it;
		for (it = used_profs.begin(); it != used_profs.end(); it++) {
			it->second = ucache.u_sum(it->first);
		}
		lp.set_obj_func(used_profs);

		return lp;
	}

	//minimize game by eliminating dominated strategies
	void minimize()
	{
		while (iterate());
		PRINTLN("Final forbidden strategies: " << fs);

		return;
	}

	//computes resulting payoffs in CE
	UProfile<T> get_payoffs(CorrelatedEquilibrium eql)
	{
		unsigned n = game.players();
		UProfile<T> up(n);
		typename CorrelatedEquilibrium::iterator it;
		for (it=eql.begin(); it != eql.end(); it++) {
			UProfile<T> act;
			while (!ucache.get(it->first, act)); //active waiting
			double prob = it->second;
			for (unsigned p=0; p<n; p++) {
				up[p] += prob * act[p];
			}
		}

		return up;
	}

	//exports reduced game
	void export_reduced(std::string filename=std::string("reduced"))
	{

		std::string fn = filename + ".game";
		std::ofstream ofs(fn.c_str());

		unsigned n = game.players();
		ofs << n << std::endl;

		for (unsigned i=0; i<n; i++) {
			ofs << fs.num_valid(i) << " ";
		}
		ofs << std::endl;

		StrategyProfile prof(n);

		//get valid profile
		bool res = valid_profile(prof, n); //valid profile with fixed nonexisting player
		if (!res) throw std::logic_error("No valid profile.");

		do {
			UProfile<T> up;
			while (!ucache.get(prof, up)); //active waiting
			for (unsigned i=0; i<n; i++) {
				ofs << std::fixed << std::setprecision(0) << up[i] << " ";
			} ofs << std::endl;
		} while (next_profile(prof, n));
		ofs.close();

		fn = filename + ".map";
		ofs.open(fn.c_str());

		ofs << n << std::endl;

		for (unsigned i=0; i<n; i++) {
			ofs << fs.num_valid(i) << " ";
		}
		ofs << std::endl;

		for (unsigned i=0; i<n; i++) {
			for (unsigned j=0; j<strategies[i]; j++) {
				if (fs.valid(i, j)) {
					ofs << j << " ";
				}
			} ofs << std::endl;
		}

		ofs.close();

		return;
	}

protected:
	//computes difference of payoffs of player "player" if he changes his
	//strategy "from" to his strategy "to" while playing strategy
	//profile "sp"; if one of needed payoffs is already being computed
	//by another thread, adds "sp" to queue "skipped"
	inline T difference(StrategyProfile sp, unsigned player, unsigned from, unsigned to, std::queue<StrategyProfile>& skipped)
	{
		bool res;
		T u_from, u_to;
		sp[player] = from;
		res = ucache.get(player, sp, u_from);
		if (!res) { //someone else computes this profile
			//save as skipped and continue
			skipped.push(sp);
			return -1;
		}
		sp[player] = to;
		res = ucache.get(player, sp, u_to);
		if (!res) { //someone else computes this profile
			//save as skipped and continue
			skipped.push(sp);
			return -1;
		}

		sp[player] = from; //not necessary

		//return to-from;
		T diff = u_from - u_to;

		return diff;
	}

	//version of difference with active waiting that is useful
	//in cases when difference should be already computed
	inline T difference(StrategyProfile sp, unsigned player, unsigned from, unsigned to)
	{
		T u_from, u_to;
		sp[player] = from;
		while (!ucache.get(player, sp, u_from));
		sp[player] = to;
		while (!ucache.get(player, sp, u_to));

		sp[player] = from; //not necessary

		//return to-from;
		T diff = u_from - u_to;

		return diff;
	}

	//moves from given VALID strategy profile "sp" to
	//next valid profile;
	//expects valid profile, doesn't check it
	inline bool next_profile(StrategyProfile& sp, unsigned player)
	{
		bool end = true;
		unsigned i = sp.size();
		while (i > 0) {
			i--;

			if (i == player) continue;

			sp[i]++;
			if (sp[i]>=strategies[i]) {
				sp[i] = 0;
				while (fs.forbidden(i, sp[i])) {
					sp[i]++;
					assert(sp[i] < strategies[i]);
				}
			} else {
				bool overflow = false;
				while (fs.forbidden(i, sp[i])) {
					sp[i]++;
					if (sp[i] >= strategies[i]) {
						assert(!overflow);
						sp[i] = 0;
						overflow = true;
					}
				}
				if (overflow) {
					continue;
				} else {
					end = false;
					break;
				}
			}
		}

		return (!end);
	}

	//checks whether strategy profile "sp" is valid; if so, does nothing,
	//else moves to next valid profile; if end of row is reached,
	//returns false
	inline bool valid_profile(StrategyProfile& sp, unsigned player)
	{
		bool end = true;

		bool overflow = false;
		for (unsigned index=sp.size(); index>0; index--) {
			unsigned i = index-1;
			//DVAR(i);
			if (i != player) {
				if (overflow) { //force move after prev. overflow
					sp[i]++;
					if (sp[i] >= strategies[i]) {
						sp[i] = 0;
						overflow = true;
					} else {
						overflow = false;
					}
				}
				while (fs.forbidden(i, sp[i])) {
					sp[i] += 1;
					if (sp[i] >= strategies[i]) {
						if (overflow) throw std::runtime_error("Double overflow.");
						sp[i] = 0;
						overflow = true;
					}
				}
			}
		}
		end = overflow;

		return !end;
	}

	//does one iteration of game reduction
	bool iterate()
	{
		//"change" is set to true if any strategy is eliminated
		bool change = false;

		static unsigned in;
		PRINT("Iteration "<<in<<": ");
		in++;
#ifdef PARALLEL
#ifdef PROCESSORS //specified number of processors
		#pragma omp parallel reduction(||:change) num_threads(PROCESSORS)
#else //default number of processors
		#pragma omp parallel reduction(||:change)
#endif
#endif
		//for players
		for (unsigned p=0; p<game.players(); p++) {
			//number of strategies of player "p"
			int pstrats = strategies[p];
				
			//for all strategies "to"
			for (int sto_=0; sto_<pstrats; sto_++) {
				//OpenMP does not like unsigned iteration variables
				unsigned sto = sto_;
				//if this strategy is forbidden, we can contunue
				if (fs.forbidden(p, sto)) continue;
				
#ifdef PARALLEL
				#pragma omp for nowait
#endif
				//for all strategies "from"
				for (int sfrom_=0; sfrom_<pstrats; sfrom_++) {
					unsigned sfrom = sfrom_;
					//if this strategy is forbidden, we can contunue
					if (fs.forbidden(p, sfrom)) continue;
					//"sfrom" and "sto" cannot be the same
					if (sfrom != sto) {
						//queue of skipped strategies
						std::queue<StrategyProfile> skipped;

						//"end" - reached end of the row
						//"step" - moved to other profile
						bool end=false, step=false;

						//profile where traversing stopped last time
						StrategyProfile prof = rows.get(p, sfrom, sto);

						//checking if profile is valid
						//(some of its strategies could
						//have been eliminated)
						if (!valid_profile(prof, p)) {
							//moved to end of the row
							fs.add(p, sfrom);
							change = true;
							continue;
						}

						//main cycle of iterating through G-matrix row
						while (!end && difference(prof, p, sfrom, sto, skipped) < 0.0) {
							step = true;
							end = !next_profile(prof, p);
						}

						//checking skipped profiles
						if (step || end) {
							while (!skipped.empty()) {
								prof = skipped.front();
								skipped.pop();
								if (difference(prof, p, sfrom, sto) >= 0.0) {
									end = false;
									break;
								}
							}
						}

						if (end) {
#ifdef OUTPUT
#ifdef PARALLEL
							#pragma omp critical(print)
#endif //PARALLEL							
							PRINT("E:" << p << "(" << sfrom << ") ");
#endif //OUTPUT
							fs.add(p, sfrom);
							change = true;
							continue;
						}

						//if we moved, store actual profile
						if (step) {
							rows.set(prof, p, sfrom, sto);
						}
					}
				}
			}
		}

		PRINT(std::endl);

		//returns whether some strategy has been eliminated
		return change;
	}
};

#endif //_G_MATRIX_HH_
