//  Copyright 2009 Stanislav Zidek
/*
 *  This file is part of CE-solver.
 *
 *  CE-solver is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CE-solver is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CE-solver.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _GAME_ADT_HH_
#define _GAME_ADT_HH_
#include "game-adt.hh"

//maximum number of player
//because of static array in StrategyProfile
#ifndef MAX_PLAYERS
#define MAX_PLAYERS 16
#endif //MAX_PLAYERS

#include <iostream> //because of StrategyProfile operator <<
#include <cassert>
#include <map>

class StrategyProfile {

friend std::ostream& operator<<(std::ostream& output, const StrategyProfile& sp);

protected:
	unsigned data[MAX_PLAYERS];
	unsigned items;

public:
	//returns size of strategy profile
	inline unsigned size() const { return items; }

	//accesses elements of strategy profile
	inline unsigned& operator[](unsigned i) { assert(i<MAX_PLAYERS); return data[i]; }
	inline const unsigned& operator[](unsigned i) const { assert(i<MAX_PLAYERS); return data[i]; }

	//constructor vith optional parameter (number of players)
	StrategyProfile(const unsigned n=MAX_PLAYERS):
		items(n)
	{
		assert(items <= MAX_PLAYERS);
		for (unsigned i=0; i<MAX_PLAYERS; i++) {
			data[i] = 0;
		}
	}

	//copy constructor
	StrategyProfile(const StrategyProfile& sp):
		items(sp.size())
	{
		assert(items <= MAX_PLAYERS);
		for (unsigned i=0; i<MAX_PLAYERS; i++) {
			data[i] = sp[i];
		}
	}

	//comparison operator, needed for indexinf std::map
	inline bool operator<(const StrategyProfile& other) const
	{
		assert(items <= MAX_PLAYERS);
		for (unsigned i=0; i<items; i++) {
			if (data[i] != other[i]) {
				return (data[i] < other[i]);
			}
		}

		return false;
	}

	//equality operator
	inline bool operator==(const StrategyProfile& other) const
	{
		assert(items <= MAX_PLAYERS);
		if (items != other.size()) return false;
		for (unsigned i=0; i<items; i++) {
			if (data[i] != other[i]) {
				return false;
			}
		}

		return true;
	}

	//assignment operator
	inline StrategyProfile& operator=(const StrategyProfile& other)
	{
		if (this == &other) return *this;
		items = other.size();
		for (unsigned i=0; i<MAX_PLAYERS; i++) {
			data[i] = other[i];
		}

		return *this;
	}
};

//stream output of StrategyProfile
std::ostream& operator<<(std::ostream& output, const StrategyProfile& sp)
{
	output << "[";
	if (sp.size() > 0) {
		output << sp[0];
		for (unsigned i=1; i<sp.size(); i++) {
			output << ", " << sp[i];
		}
	}
	output << "]";

	return output;
}

//class for storing payoffs of players in one strategy profile
template <class T>
class UProfile {

template <class U>
friend std::ostream& operator<<(std::ostream& output, const UProfile<U>& up);

private:
	T data[MAX_PLAYERS];
	unsigned items;
public:
	//returns size of payoff profile
	inline unsigned size() const { return items; }
	//accesses elements of payoff profile
	inline T& operator[](unsigned i) { /*assert(i<items);*/ return data[i]; }
	inline const T& operator[](unsigned i) const { /*assert(i<items);*/ return data[i]; }

	//constructor with optional parameter (number of players)
	UProfile(unsigned n=MAX_PLAYERS):
		items(n)
	{
		for (unsigned i=0; i<MAX_PLAYERS; i++) {
			data[i] = 0.0;
		}
	}

	//assignment operator
	inline UProfile& operator=(const UProfile& other)
	{
		if (this == &other) return *this;

		items = other.size();
		for (unsigned i=0; i<items; i++) {
			data[i] = other[i];
		}

		return *this;
	}
};

//stream output of UProfile
template <class T>
std::ostream& operator<<(std::ostream& output, const UProfile<T>& up)
{
	output << "[";
	if (up.size() > 0) {
		output << up[0];
		for (unsigned i=1; i<up.size(); i++) {
			output << ", " << up[i];
		}
	}
	output << "]";

	return output;
}

//correlated equilibrium - probability dostribution over strategy profiles
typedef std::map<StrategyProfile, double> CorrelatedEquilibrium;

//class representing game
template <class T>
class Game {
public:
	//returns number of players
	virtual unsigned players() const =0;
	//returns number of players' strategies
	virtual StrategyProfile strategies() const =0;
	//returns players' payoffs for given strategy profile
	virtual UProfile<T> get(const StrategyProfile&) const =0;

	/* 
	 * There is no need to implement next two methods, but may make sense
	 * for performacne reasons.
	 */

	//returns number of strategies of given player
	virtual unsigned strategies(unsigned player) const
	{
		return (strategies())[player];
	}

	//returns payoff of given player in given strategy profile
	virtual T get(const unsigned player, const StrategyProfile& sp) const
	{
		return (get(sp))[player];
	}
};

#endif //_GAME_ADT_HH_
