//  Copyright 2009 Stanislav Zidek
/*
 *  This file is part of CE-solver.
 *
 *  CE-solver is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CE-solver is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CE-solver.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

//Interface to GLPK for solving LP problems

#ifndef _GLPK_SOLVER_HH_
#define _GLPK_SOLVER_HH_
#include "glpk-solver.hh"

#include "game-adt.hh"
#include "ce-lp.hh"

#include <glpk.h>

#include <vector>
#include <map>

#include <cstdio>

template <class T>
class GLPKSolver {
protected:
	const CorrelatedLP<T> &clp;
	std::vector<double> solution;
	double z; //objective value
public:
	GLPKSolver(const CorrelatedLP<T>& corr_lp):
		clp(corr_lp)
	{}

	CorrelatedEquilibrium solve()
	{
		create();

		return clp.translate(solution);
	}

protected:
	//creates and solves problem from CorrelatedLP
	void create()
	{
		unsigned rows = clp.get_rows();
		unsigned cols = clp.get_cols();
		unsigned items = clp.get_items();
		unsigned size = items + cols + 1;

		int* ia = new int[size];
		int* ja = new int[size];
		double* ar = new double[size];

		//GLPK problem
		glp_prob *lp;
		lp = glp_create_prob();

		glp_set_obj_dir(lp, GLP_MAX);
		glp_add_cols(lp, cols);
		std::vector<T> obj_func = clp.get_obj_func();
		DPRINT("OFS: " << obj_func.size() << ", Cols: " << cols);
		assert(obj_func.size()==cols);
		for (unsigned i=0; i<obj_func.size(); i++) {
			DPRINT("Objective function[" << i+1 << "] = " << obj_func[i]);
			glp_set_obj_coef(lp, i+1, obj_func[i]);
			glp_set_col_bnds(lp, i+1, GLP_DB, 0.0, 1.0);
		}

		//additional row, not in CorrelatedLP
		glp_add_rows(lp, rows+1); //row with all ones (sum of probabilities)
		for (unsigned i=1; i<=rows; i++) {
			glp_set_row_bnds(lp, i, GLP_LO, 0.0, 0.0);
		}
		//row with all ones
		glp_set_row_bnds(lp, rows+1, GLP_FX, 1.0, 1.0);

		//values from CorrelatedLP
		for (unsigned i=0; i<items; i++) {
			unsigned r,c;
			T v;
			clp.get(i, r, c, v);
			r++; c++;
			ia[i+1] = r;
			ja[i+1] = c;
			ar[i+1] = v;
		}
		//last row (sum of probabilities)
		for (unsigned i=0; i<cols; i++) {
			unsigned index = items + 1 + i;
			ia[index] = rows+1;
			ja[index] = i+1;
			ar[index] = 1.0;
		}
		for (unsigned i=1; i<size; i++) {
			DPRINT(ia[i]<<","<<ja[i]<<"="<<ar[i]);
		}

		glp_load_matrix(lp, size-1, ia, ja, ar);

#ifdef DEBUG //print problem
		printf("LP problem from GLPK:\n");
		printf("%d %d %d %d %d\n",GLP_FR,GLP_UP,GLP_LO,GLP_DB,GLP_FX);
		printf("Rows: %d\n", glp_get_num_rows(lp));
		int m = glp_get_num_rows(lp);
		int n = glp_get_num_cols(lp);
		printf("# %d %d\n", n, m);
		int t = glp_get_obj_dir(lp);
		if (t==GLP_MIN) printf("0");
		else printf("1");
		for (int j=1; j<=n; j++) {
			double coef = glp_get_obj_coef(lp, j);
			printf(" %g", coef);
		}
		printf("\n");
		double *arr = new double[n];
		for (int i=1; i<=glp_get_num_rows(lp); i++) {
			int type = glp_get_row_type(lp, i);
			double ub = lpx_get_row_ub(lp, i);
			double lb = lpx_get_row_lb(lp, i);
			if (type==GLP_FX) printf("= %g", lb);
			else if (type==GLP_LO) printf("> %g", lb);
			else printf("< %g", ub);
			int *ind= new int[n+1];
			double *val= new double[n+1];
			int len = glp_get_mat_row(lp, i, ind, val);
			for (int j=0; j<n; j++) {
				arr[j] = 0;
			}
			for (int j=0; j<len; j++) {
				arr[ind[j+1]-1] = val[j+1];
			}
			for (int j=0; j<n; j++) {
				printf(" %g", arr[j]);
			}
			/*
			int l=1;
			for (unsigned j=1; j<=glp_get_num_cols(lp); j++) {
				printf("%d:%g; ", ind[j], val[j]);
			}
			for (unsigned j=1; j<=glp_get_num_cols(lp); j++) {
				if (l<=n && ind[l]==j) {
					printf(" %g", val[j]);
					l++;
					continue;
				}
				printf(" 0");
			}
			*/
			printf("\n");
			delete [] ind;
			delete [] val;
			//printf("%u: %d %g %g\n", i, type, lb, ub);
		}
		delete [] arr;
		printf("Cols: %d\n", glp_get_num_cols(lp));
		for (int i=1; i<=glp_get_num_cols(lp); i++) {
			int type = glp_get_col_type(lp, i);
			double ub = lpx_get_col_ub(lp, i);
			double lb = lpx_get_col_lb(lp, i);
			printf("C%u: %d %g %g\n",i, type, lb, ub);
		}
#endif
		//let's shut GLPK up
		glp_smcp parm; //GLPK control parameters
		glp_init_smcp(&parm);
#ifndef OUTPUT
		parm.msg_lev = GLP_MSG_OFF;
#endif

		//call simplex method
		glp_simplex(lp, &parm);

		z = glp_get_obj_val(lp);
		for (unsigned i=1; i<=cols; i++) {
			solution.push_back(glp_get_col_prim(lp, i));
			DPRINT("Var: "<<i<<":"<<glp_get_col_prim(lp,i));
		}

		glp_delete_prob(lp);
		delete [] ia;
		delete [] ja;
		delete [] ar;

		return;
	}
};

#endif //_GLPK_SOLVER_HH_
