//  Copyright 2009 Stanislav Zidek
/*
 *  This file is part of CE-solver.
 *
 *  CE-solver is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CE-solver is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with CE-solver.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

//Payoff cache

#ifndef _UCACHE_HH_
#define _UCACHE_HH_
#include "ucache.hh"

#include "debug.hh"
#include "game-adt.hh"

#ifdef PARALLEL
#include <omp.h>
#endif

#include <algorithm>
#include <set>
#include <vector>
#include <map>
#include <stdexcept>

//payoff cache declaration - implementation depends on B
template <bool B, class T> class UCache;

//fake cache that calls game model every time
template <class T>
class UCache<false, T> {
private:
	const Game<T>& game;
public:
	UCache(const Game<T>& g):
		game(g)
	{}

	//gets payoff profile for given strategy profile
	inline bool get(const StrategyProfile& sp, UProfile<T>& up)
	{
		up = game.get(sp);
		return true;
	}

	//gets payoff profile for given strategy profile
	inline UProfile<T> get(const StrategyProfile& sp)
	{
		return game.get(sp);
	}

	//gets payoff of given player for given strategy profile
	inline bool get(unsigned player, const StrategyProfile& sp, T& payoff)
	{
		payoff = game.get(player, sp);
		return true;
	}

	//gets payoff of given player for given strategy profile
	inline T get(unsigned player, const StrategyProfile& sp)
	{
		return game.get(player, sp);
	}

	//returns sum of players' payoffs in given strategy profile
	T u_sum(const StrategyProfile& sp)
	{
		UProfile<T> up;
		get(sp, up); //no need to test return value in this specialization
		DVAR(up);
		T sum = 0.0;
		for (unsigned i=0; i<sp.size(); i++) {
			sum += up[i];
		}

		return sum;
	}
};

//real ucache
template <class T>
class UCache<true, T> {
private:
	const Game<T>& game;
	std::map<StrategyProfile, UProfile<T> > data;
	std::set<StrategyProfile> act_computing;
public:
	UCache(const Game<T>& g):
		game(g)
	{}

	//gets payoff profile for given strategy profile
	inline bool get(const StrategyProfile& sp, UProfile<T>& up)
	{
		unsigned n = game.players();
		assert(n == sp.size());
		//typename std::set<StrategyProfile>::iterator ac_iter;
		bool res=false, computed;
		
#ifdef PARALLEL
#pragma omp critical(ucache)
{
#endif
		if (!act_computing.count(sp)) {
			//payoffs are NOT being computed right now
			res = true; //we will return payoffs
			typename std::map<StrategyProfile, UProfile<T> >::iterator it;
			it = data.find(sp);
			computed = (it != data.end());
			if (computed) {
				//payoffs are already computed
				up = it->second;
			} else {
				//we are going to compute payoffs
				act_computing.insert(sp);
			}
		}
#ifdef PARALLEL
}
#endif

		if (!res) {
			//someone else is computing requested payoffs
			return false;
		}

		if (!computed) {
			//we must compute payoffs
			up = game.get(sp);
#ifdef PARALLEL
#pragma omp critical(ucache)
{
#endif
			//insert into cache
			data[sp] = up;
			//remove from list of currently computed payoffs
			act_computing.erase(sp);
#ifdef PARALLEL
}
#endif
		}


		return res;
	}

	//gets payoff profile for given strategy profile
	UProfile<T> get(const StrategyProfile& sp)
	{
		UProfile<T> up;
		while (!get(sp, up));

		return up;
	}

	//gets payoff of given player for given strategy profile
	inline bool get(unsigned player, const StrategyProfile& sp, T& payoff)
	{
		UProfile<T> up;
		bool res = get(sp, up);
		payoff = up[player];

		return res;
	}

	//gets payoff of given player for given strategy profile
	T get(unsigned player, const StrategyProfile& sp)
	{
		T payoff;
		while (!get(player, sp, payoff));

		return payoff;
	}

	//returns sum of players' payoffs in given strategy profile
	//we suppose payoffs for sp are not being computed
	T u_sum(const StrategyProfile& sp)
	{
		UProfile<T> up = get(sp);
		//bool res = get(sp, up);
		//if (!res) throw std::logic_error("Payoffs in u_sum are being computed.");

		DPRINT(sp.size() << " || " << up.size());
		assert(sp.size() == up.size());
		T sum = 0.0;
		for (unsigned i=0; i<sp.size(); i++) {
			sum += up[i];
		}

		return sum;
	}
};

#endif //_UCACHE_HH_
